from typing import Callable
from typing import List
from typing import Tuple

from chocs.errors.not_found_error import NotFoundError
from .route import Route


class Router:
    def __init__(self):
        self._routes: List[Tuple[Route, Callable]] = []

    def append(self, route: Route, handler: Callable) -> None:
        assert isinstance(route, Route), "Passed route must be instance of Route"

        self._routes.append((route, handler))
        self._routes.sort(key=lambda r: r[0].wildcard)

    def match(self, uri: str) -> Tuple[Route, Callable]:
        for route in self._routes:
            if route[0].match(uri):
                return route

        raise NotFoundError(f"Could not match any resource matching {uri} uri")


__all__ = ["Router"]



import json
from io import BytesIO
from json.decoder import JSONDecodeError

from .body import RequestBody


class JsonBody(RequestBody):
    @classmethod
    def from_wsgi(cls, wsgi_input: BytesIO, encoding: str = "utf8") -> "JsonBody":
        wsgi_input.seek(0)
        decoded_input = wsgi_input.read().decode(encoding)

        try:
            body = json.loads(decoded_input)
        except JSONDecodeError:
            body = {}

        instance = cls(body)

        return instance


__all__ = ["JsonBody"]



class HttpError(Exception):
    status_code: int = 500
    http_message = "Internal Server Error"

    def __init__(self, http_message: str, status_code: int = 500):
        self.status_code = status_code
        self.http_message = http_message
        super().__init__(http_message)


__all__ = ["HttpError"]



from .chocs import Application
from .chocs import router
from .chocs import serve
from .cookie_jar import Cookie
from .cookie_jar import CookieJar
from .cookie_jar import CookieSameSitePolicy
from .headers import Headers
from .http_method import HttpMethod
from .http_request import HttpRequest
from .http_response import HttpResponse
from .http_status import HttpStatus
from .query_string import QueryString
from .query_string import parse_qs



from io import BytesIO

from chocs.query_string import parse_qs
from .body import RequestBody


class FormBody(RequestBody):
    @classmethod
    def from_wsgi(cls, wsgi_input: BytesIO, encoding: str = "utf8") -> "FormBody":
        wsgi_input.seek(0)
        decoded_input = wsgi_input.read().decode(encoding)
        fields = parse_qs(decoded_input)
        instance = cls()

        for name, value in fields.items():
            instance[name] = value

        return instance


__all__ = ["FormBody"]



from cgi import parse_header
from enum import Enum
from io import BytesIO
from tempfile import TemporaryFile
from typing import Any
from typing import Dict
from typing import IO
from typing import Text
from typing import Tuple
from typing import Union

from .form_body import FormBody


class ParserState(Enum):
    PART_BOUNDARY = 0
    CONTENT_DISPOSITION = 1
    CONTENT_TYPE = 2
    CONTENT_HEADER = 3
    CONTENT_DATA = 4
    END = 5


class MultipartBody(FormBody):
    @staticmethod
    def from_wsgi(
        wsgi_input: BytesIO, encoding: str = "utf8", boundary: str = ""
    ) -> "MultipartBody":
        assert boundary, (
            "%s.from_wsgi requires boundary parameters." % MultipartBody.__name__
        )
        wsgi_input.seek(0)
        return _parse_multipart_data(wsgi_input.read(), boundary, encoding)


def _parse_multipart_data(
    data: bytes, boundary: str, encoding: str = "utf8"
) -> MultipartBody:
    state = ParserState.PART_BOUNDARY
    prev_byte = None
    cursor = 0
    boundary_length = len(boundary)
    string_buffer = ""
    body = MultipartBody()

    def _append_content_to_body(
        raw_content_disposition: str, _content_type: str, _content_data: bytes
    ) -> None:
        parsed_content_disposition: Tuple[str, Dict[str, str]] = parse_header(
            raw_content_disposition[20:]
        )
        if "filename" in parsed_content_disposition[1]:
            tmp_file = TemporaryFile()
            tmp_file.write(_content_data)
            tmp_file.seek(0)
            body[parsed_content_disposition[1]["name"]] = UploadedFile(
                tmp_file,
                _content_type[14:].lower(),
                parsed_content_disposition[1]["filename"],
            )
        else:
            body[parsed_content_disposition[1]["name"]] = _content_data.decode(encoding)

    content_disposition = ""
    content_type = ""
    content_cursor = 0

    for code in data:
        line_break = code == 0x0A and prev_byte == 0x0D
        new_line_char = code == 0x0A or code == 0x0D

        if not new_line_char:
            string_buffer = string_buffer + chr(code)

        if state is ParserState.PART_BOUNDARY and line_break:
            if string_buffer == "--" + boundary:
                string_buffer = ""
                state = ParserState.CONTENT_DISPOSITION
            else:
                raise IOError(
                    "Could not parse message body, body is malformed or incorrect boundary was passed."
                )

        elif state is ParserState.CONTENT_DISPOSITION and line_break:
            content_disposition = string_buffer
            string_buffer = ""
            state = ParserState.CONTENT_TYPE
            if (
                data[cursor + 1 : cursor + 13].decode(encoding).lower()
                != "content-type"
            ):
                state = ParserState.CONTENT_HEADER
        elif state is ParserState.CONTENT_TYPE and line_break:
            content_type = string_buffer
            state = ParserState.CONTENT_HEADER
            string_buffer = ""
        elif state is ParserState.CONTENT_HEADER and line_break:
            string_buffer = ""
            state = ParserState.CONTENT_DATA
            content_cursor = cursor
        if state is ParserState.CONTENT_DATA:
            if line_break and string_buffer == "--" + boundary:
                content_data = data[content_cursor + 1 : cursor - (boundary_length + 5)]
                _append_content_to_body(content_disposition, content_type, content_data)
                content_type = ""
                state = ParserState.CONTENT_DISPOSITION

            if line_break and string_buffer == "--" + boundary + "--":
                content_data = data[content_cursor + 1 : cursor - (boundary_length + 7)]
                _append_content_to_body(content_disposition, content_type, content_data)
                content_type = ""
                state = ParserState.END

            if line_break:
                string_buffer = ""

        prev_byte = code
        cursor += 1

    return body


class UploadedFile:
    """
    Proxy class for uploaded file
    """

    def __init__(self, file: IO[Any], mimetype: str, filename: str):
        self.file = file
        self.mimetype = mimetype
        self.filename = filename
        self.length = 0
        self._str = ""

    def read(self) -> bytes:
        return self.file.read()

    def seek(self, offset: int) -> int:
        return self.file.seek(offset)

    def close(self) -> None:
        self.file.close()

    def save(self, path: str) -> IO[Any]:
        if self.file.closed:
            raise ValueError(f"Cannot save to file {path} of closed stream.")
        with open(path, "wb") as file:
            self.seek(0)
            file.write(self.read())

        return file

    def __float__(self) -> None:
        raise ValueError(
            f"Cannot convert instance of {TemporaryFile.__name__} to float"
        )

    def __int__(self) -> None:
        raise ValueError(f"Cannot convert instance of {TemporaryFile.__name__} to int")

    def __len__(self) -> int:
        if not self.length:
            bytes = self.read()
            self.seek(0)
            self.length = len(bytes)
        return self.length

    def __bool__(self) -> bool:
        return len(self) > 0

    def __str__(self) -> str:

        if not self._str:
            self._str = self.read().decode()

        return self._str

    def __enter__(self) -> IO[Any]:
        return self.file

    def __exit__(self, *args: Any) -> None:
        self.close()


__all__ = ["MultipartBody", "UploadedFile"]



from abc import ABC
from abc import abstractmethod

from chocs.http_request import HttpRequest
from chocs.http_response import HttpResponse
from .middleware_handler import MiddlewareHandler


class Middleware(ABC):
    @abstractmethod
    def handle(self, request: HttpRequest, next: MiddlewareHandler) -> HttpResponse:
        pass


__all__ = ["Middleware"]



from typing import Callable
from typing import Dict

from chocs.errors import HttpError
from chocs.http_method import HttpMethod
from chocs.http_request import HttpRequest
from chocs.http_response import HttpResponse
from chocs.middleware import Middleware
from chocs.middleware.middleware_handler import MiddlewareHandler
from chocs.routing.route import Route
from chocs.routing.router import Router


class RoutingMiddleware(Middleware):
    def __init__(self) -> None:
        self.methods: Dict[HttpMethod, Router] = {key: Router() for key in HttpMethod}

    def handle(self, request: HttpRequest, next: MiddlewareHandler) -> HttpResponse:
        try:
            route, controller = self.methods[request.method].match(
                request.uri
            )  # type: Route, Callable

            request.attributes = route.attributes
            response: HttpResponse = controller(request)

            return response
        except HttpError as error:
            return HttpResponse(status=error.status_code, body=error.http_message)


__all__ = ["RoutingMiddleware"]



from collections import UserDict
from typing import Any
from typing import Optional


class RequestBody(UserDict):
    def get(self, name: str, default: Optional[Any] = None) -> Any:
        if name in self:
            return self[name]

        return default


__all__ = ["RequestBody"]
